(in-package :pong.client)

(defparameter *socket* nil)
(defparameter *stream* nil)
(defparameter *game* nil)
(defparameter *lock* (bt:make-lock))

(defclass online-game (g:online-game)
  ((keyboard :initform (make-hash-table) :reader online-game-keyboard)
   (paddle :initarg :paddle :initform :left :reader online-game-paddle)))

(defun apply-actions (game)
  (let* ((state-timestamp (g:state-timestamp (g:game-state game)))
	 (last-time state-timestamp))
    (loop for action in (g:actions game)
	  when (>= (g:action-timestamp action) state-timestamp)
	    do (g:on-update game (- (g:action-timestamp action) last-time))
	       (setf last-time (g:action-timestamp action))
	       (g:apply-action action game))
    (let ((current-time (g:get-time)))
      (g:on-update game (- current-time last-time))
      (setf (g:state-timestamp (g:game-state game)) current-time))
    (setf (g:actions game) nil)))

(defmethod apply-message ((msg g:state-updated-message))
  (setf (g:game-state *game*) (g:game-state msg))
  (apply-actions *game*))

(defmethod apply-message ((msg g:game-joined-message))
  (setf (gui:visible (game-scene-popup *scene*)) nil))

(defmethod apply-message ((msg g:player-left-message))
  (setf (gui:visible (game-scene-popup *scene*)) t)
  (setf (gui:text (game-scene-popup-text *scene*))
	(format nil "Player ~a left the game.~%Waiting for player..." (g:player-name msg))))

(defun game-thread ()
  (when (usocket:wait-for-input *socket* :ready-only t :timeout 5)
    (bt:with-lock-held (*lock*)
      (when *game*
	(let ((msg (cpk:decode-stream *stream*)))
	  (apply-message msg)))))
  (when *game*
    (game-thread)))

(defun start-online-game% (game)
  (setf *game* game)
  (set-scene (open-game *game*))
  (bt:make-thread #'game-thread))

(defmacro start-online-game (&rest args)
  `(start-online-game% (make-instance 'online-game ,@args)))

(defun open-socket ()
  (setf *socket* (usocket:socket-connect "pariatech.com" 54321 :element-type 'unsigned-byte))
  (setf *stream* (usocket:socket-stream *socket*)))

(defun write-to-server (data)
  (cpk:encode data :stream *stream*)
  (force-output *stream*))

(defmethod handle-join-online-game-response ((response g:game-joined-message) code)
  (let ((player-paddle (g:player-paddle response))
	(game-state (g:game-state response)))
    (start-online-game :code code
		       :paddle player-paddle
		       :state game-state)
    :success))

(defmethod handle-join-online-game-response ((response g:game-full) code)
  (declare (ignore code))
  :game-full)

(defmethod handle-join-online-game-response ((response g:game-does-not-exist) code)
  (declare (ignore code))
  :game-does-not-exist)

(defun join-online-game (name code)
  (open-socket)
  (write-to-server (make-instance 'g:join-game-message
				  :player-name name
				  :game-code code))
  (usocket:wait-for-input *socket*) ;; wait for response
  (let ((response (cpk:decode-stream *stream*)))
    (handle-join-online-game-response response code)))

(defun create-online-game (name)
  (open-socket)
  (write-to-server (make-instance 'g:create-game-message :player-name name))
  (usocket:wait-for-input *socket*) ;; wait for response

  ;; should be game-created-message
  (let* ((response (cpk:decode-stream *stream*))
	 (game-code (g:game-code response))
	 (player-paddle (g:player-paddle response))
	 (game-state (g:game-state response)))
    (start-online-game :code game-code
		       :paddle player-paddle
		       :state game-state))

  (setf (gui:visible (game-scene-popup *scene*)) t))

(defmethod g:on-init ((game online-game)))

(defun is-first-key-down-p (key keyboard)
  (and (r:is-key-down key) (not (gethash key keyboard))))

(defun is-first-key-released-p (key keyboard)
  (and (r:is-key-released key) (gethash key keyboard)))

(defun set-key-down (key keyboard)
  (setf (gethash key keyboard) (r:is-key-down key)))

(defun handle-action (action game)
  (bt:with-lock-held (*lock*)
    (g:apply-action action game)
    (push action (g:actions game)))
  (write-to-server (make-instance 'g:game-action-message
				  :game-code (g:game-code game)
				  :game-action action)))

(defmethod g:on-update :before ((game online-game) timelapse)
  (with-slots (keyboard paddle) game
    (when (or (is-first-key-down-p r:+key-s+ keyboard)
	      (is-first-key-down-p r:+key-down+ keyboard))
      (set-key-down r:+key-s+ keyboard)
      (set-key-down r:+key-down+ keyboard)
      (let ((action (make-instance 'g:start-down-action :paddle paddle)))
	(handle-action action game)))

    (when (or (is-first-key-released-p r:+key-s+ keyboard)
	      (is-first-key-released-p r:+key-down+ keyboard))
      (set-key-down r:+key-s+ keyboard)
      (set-key-down r:+key-down+ keyboard)
      (let ((action (make-instance 'g:stop-down-action :paddle paddle)))
	(handle-action action game)))

    (when (or (is-first-key-down-p r:+key-w+ keyboard)
	      (is-first-key-down-p r:+key-up+ keyboard))
      (set-key-down r:+key-w+ keyboard)
      (set-key-down r:+key-up+ keyboard)
      (let ((action (make-instance 'g:start-up-action :paddle paddle)))
	(handle-action action game)))

    (when (or (is-first-key-released-p r:+key-w+ keyboard)
	      (is-first-key-released-p r:+key-up+ keyboard))
      (set-key-down r:+key-w+ keyboard)
      (set-key-down r:+key-up+ keyboard)
      (let ((action (make-instance 'g:stop-up-action :paddle paddle)))
	(handle-action action game)))))

(defmethod g:on-quit ((game online-game))
  (usocket:socket-close *socket*)
  (setf *socket* nil)
  (setf *stream* nil)
  (setf *game* nil))
