(in-package :pong.game)

(defconstant +paddle-speed+ 0.4)
(defconstant +paddle-height+ 0.1)
(defconstant +paddle-width+ 0.02)
(defconstant +ball-radius+ 0.01)
(defconstant +max-launch-angle+ (/ pi 4)) ; 45°
(defconstant +ball-speed+ 0.5)

(defun get-time ()
  (multiple-value-bind (sec nsec) (sb-ext:get-time-of-day)
    (+ sec (/ nsec 1000000))))

(defstruct paddle
  (y  0.0 :type float)
  (vy 0.0 :type float))

(defstruct ball
  (x 0.0 :type float)
  (y 0.0 :type float)
  (vx 0.0 :type float)
  (vy 0.0 :type float))

(defclass state ()
  ((timestamp    :initarg :timestamp
	         :initform (get-time)
	         :accessor state-timestamp)
   (left-paddle  :initarg :left-paddle
		 :initform (make-paddle :y 0.5)
		 :accessor state-left-paddle)
   (right-paddle :initarg :right-paddle
		 :initform (make-paddle :y 0.5)
		 :accessor state-right-paddle)
   (ball         :initarg :ball
		 :initform (make-ball :x 0.5 :y 0.5)
		 :accessor state-ball)
   (bounces      :initarg :bounces
		 :initform 0
		 :accessor state-bounces)
   (left-score   :initarg :left-score
		 :initform 0
		 :accessor state-left-score)
   (right-score  :initarg :right-score
		 :initform 0
		 :accessor state-right-score)
   (left-player  :initarg :left-player
		 :initform nil
		 :accessor state-left-player)
   (right-player :initarg :right-player
		 :initform nil
		 :accessor state-right-player)
   (paused       :initarg :paused
		 :initform nil
		 :reader state-paused)))

(cpk:defencoding state timestamp left-paddle right-paddle ball bounces left-score right-score left-player right-player paused)
(cpk:defencoding paddle y vy)
(cpk:defencoding ball x y vx vy)

(defun random-game-code (length)
  (let ((chars "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789")
        (password (make-string length)))
    (dotimes (i length)
      (setf (aref password i) (aref chars (random (length chars)))))
    password))

(defclass game ()
  ((state :initarg :state
	  :initform (make-instance 'state)
	  :accessor game-state)))

(defclass online-game (game)
  ((code :initarg :code
	 :initform (random-game-code 6)
	 :reader game-code)
   (actions :initarg :actions :initform nil :accessor actions)))

(defgeneric on-update (game timelapse))
(defgeneric on-init (game))
(defgeneric on-quit (game))

(defun update-paddle (paddle timelapse)
  ;; (format t "timelapse: ~a~%" (float timelapse))
 (setf (paddle-y paddle)
       (min (max (+ (paddle-y paddle) (* (paddle-vy paddle) timelapse))
		 (/ +paddle-height+ 2))
	    (- 1 (/ +paddle-height+ 2)))))

(defun point-in-rect-p (px py rx ry rw rh)
  (and (> px rx)
       (<  px (+ rx rw))
       (> py ry)
       (<  py (+ ry rh))))

(defun get-ball-paddle-collision (ball paddle paddle-x)
  (let* ((ball-x (ball-x ball))
	 (ball-y (ball-y ball))
	 (paddle-y (paddle-y paddle)))
    (when (or (point-in-rect-p (- ball-x +ball-radius+)
			       (- ball-y +ball-radius+)
			       paddle-x
			       (- paddle-y (/ +paddle-height+ 2))
			       +paddle-width+
			       +paddle-height+)
	      (point-in-rect-p (- ball-x +ball-radius+)
			       (+ ball-y +ball-radius+)
			       paddle-x
			       (- paddle-y (/ +paddle-height+ 2))
			       +paddle-width+
			       +paddle-height+)
	      (point-in-rect-p (+ ball-x +ball-radius+)
			       (- ball-y +ball-radius+)
			       paddle-x
			       (- paddle-y (/ +paddle-height+ 2))
			       +paddle-width+
			       +paddle-height+)
	      (point-in-rect-p (+ ball-x +ball-radius+)
			       (+ ball-y +ball-radius+)
			       paddle-x
			       (- paddle-y (/ +paddle-height+ 2))
			       +paddle-width+
			       +paddle-height+))
      (- ball-y paddle-y))))

(defmethod launch-ball ((ball ball) bounces rad x y)
  (let* ((v (v:v* (v:vec (cos rad) (sin rad)) (* (log (+ bounces 2)) +ball-speed+))))
    (setf (ball-x ball) x)
    (setf (ball-y ball) y)
    (setf (ball-vx ball) (v:vx v))
    (setf (ball-vy ball) (v:vy v))))

(defmethod random-launch-ball ((state state))
  (launch-ball (state-ball state)
	       (setf (state-bounces state) 0)
	       (+ (random +max-launch-angle+) (* (random 2) pi))
	       0.5 
	       (random 1.0)))

(defun update-ball (state ball left-paddle right-paddle timelapse)
  (with-slots (x y vx vy) ball
    (incf x (* vx timelapse))
    (incf y (* vy timelapse))
    (let ((ball-left-paddle-collision
	    (get-ball-paddle-collision ball left-paddle 0.0))
	  (ball-right-paddle-collision
	    (get-ball-paddle-collision ball right-paddle (- 1.0 +paddle-width+))))
    (cond
      (ball-left-paddle-collision
       (launch-ball ball
		    (incf (state-bounces state))
		    (* (/ ball-left-paddle-collision
			  (+ (/ +paddle-height+ 2) +ball-radius+))
		       +max-launch-angle+)
		    (+ +paddle-width+ +ball-radius+)
		    y))
      (ball-right-paddle-collision
       (launch-ball ball
		    (incf (state-bounces state))
		    (- pi
		       (* (/ ball-right-paddle-collision
			     (+ (/ +paddle-height+ 2) +ball-radius+))
			  (/ +max-launch-angle+ 2)))
		    (- 1.0 +paddle-width+ +ball-radius+)
		    y))
      ((minusp (+ x +ball-radius+))
       (incf (state-right-score state))
       (random-launch-ball state))
      ((> (- x +ball-radius+) 1)
       (incf (state-left-score state))
       (random-launch-ball state))
      ((minusp (- y +ball-radius+))
       (setf y (+ (- y) (* +ball-radius+ 2)))
       (setf vy (- vy)))
      ((> (+ y +ball-radius+) 1)
       (setf y (- 2 y (* +ball-radius+ 2)))
       (setf vy (- vy)))))))

(defmethod on-update ((game game) timelapse)
  (let* ((state (game-state game))
	 (left-paddle (state-left-paddle state))
	 (right-paddle (state-right-paddle state))
	 (left-player (state-left-player state))
	 (right-player (state-right-player state))
	 (ball (state-ball state))
	 (paused (state-paused state)))
    (unless (or paused (not (and left-player right-player)))
      (update-paddle left-paddle timelapse)
      (update-paddle right-paddle timelapse)
      (update-ball state ball left-paddle right-paddle timelapse))))

(defmethod pause ((game game))
  (setf (slot-value (game-state game) 'paused) t))

(defmethod resume ((game game))
  (setf (slot-value (game-state game) 'paused) nil))

