(in-package :pong.server)

(defclass online-game (g:online-game)
  ((lock :initarg :lock :initform (bt:make-lock) :reader online-game-lock)
   (left-player-connection :initarg :left-player-connection
			   :initform nil
			   :accessor online-game-left-player-connection)
   (right-player-connection :initarg :right-player-connection
			    :initform nil
			    :accessor online-game-right-player-connection)
   (connections :initarg :connections :initform nil :accessor online-game-connections)))

(defparameter *server* nil)
(defparameter *server-running* nil)

(defparameter *connections* (make-hash-table))
(defparameter *games* (make-hash-table :test 'equal))

(defun get-connections ()
   (loop for key being the hash-keys of *connections* collect key))

(defmacro loop-while-server-running (&body body)
  `(loop while *server-running* do (progn ,@body)))

(defun send-data-to-client (data client)
  (let ((stream (usocket:socket-stream client)))
    (when stream
      (cpk:encode data :stream stream)
      (force-output stream))))

(defmacro send-message-to-client (client message &rest args)
  `(send-data-to-client (make-instance ',message ,@args) ,client))

(defun update-game (game)
  (bt:with-lock-held ((online-game-lock game))
    (let* ((actions (g:actions game))
	   (connections (online-game-connections game))
	   (state (g:game-state game))
	   (last-time (g:state-timestamp state)))
      (loop for action in (sort actions '< :key #'g:action-timestamp)
	    do (g:on-update game (print (float (- (g:action-timestamp action) last-time))))
	       (setf last-time (g:action-timestamp action))
	       (g:apply-action action game))

      (let ((current-time (g:get-time)))
	(g:on-update game (- current-time last-time))
	(setf (g:state-timestamp state) current-time))

      (setf (g:actions game) nil)

      (loop for con in connections
	    do (handler-case (send-message-to-client con g:state-updated-message :game-state state)
		 (t ()
		   (format t "Socket closed???~%")
		   (close-connection con)))
	       ))))

(defun update-games ()
  (loop-while-server-running
    (loop for game being the hash-values of *games*
	  do (update-game game))
    (sleep .100)))

(defun handle-client (client)
  (let* ((stream (usocket:socket-stream client))
	 (action (cpk:decode-stream stream)))
    (handle-message action stream client)))

(defun open-server-socket (port)
  (usocket:socket-listen "0.0.0.0" port :reuse-address t :element-type 'unsigned-byte))

(defun wait-for-sockets-ready (connections)
  (usocket:wait-for-input connections :ready-only t :timeout 5))

(defun close-connection (connection)
  (usocket:socket-close connection) 
  (let* ((game (gethash connection *connections*))
	 (state (and game (g:game-state game))))
    (when game
      (setf (online-game-connections game) (remove connection (online-game-connections game)))
      (cond ((eq (online-game-left-player-connection game) connection)
	     (setf (online-game-left-player-connection game) nil)
	     (let ((player (g:state-left-player state)))
	       (setf (g:state-left-player state) nil)
	       (when (online-game-right-player-connection game)
		 (send-message-to-client (online-game-right-player-connection game)
					 g:player-left-message
					 :player-name player
					 :game-state state))))
	    ((eq (online-game-right-player-connection game) connection)
	     (setf (online-game-right-player-connection game) nil)
	     (let ((player (g:state-right-player state)))
	       (setf (g:state-right-player state) nil)
	       (when (online-game-left-player-connection game)
		 (send-message-to-client (online-game-left-player-connection game)
					 g:player-left-message
					 :player-name player
					 :game-state state)))))
      
      (unless (online-game-connections game)
	(remhash (g:game-code game) *games*))))
  (remhash connection *connections*))

(defun start-server (port)
  (unless *server-running*
    (setf *server-running* t)

    (bt:make-thread #'update-games)
    (let* ((server-socket (open-server-socket port)))
      (unwind-protect
	   (loop-while-server-running
	    (loop for ready in (wait-for-sockets-ready (cons server-socket (get-connections)))
		  while *server-running*
		  do (if (usocket:stream-server-usocket-p ready)
			 (setf (gethash (usocket:socket-accept ready) *connections*) nil)
			 (handler-case (handle-client ready)
			   (stream-error ()
			     (format t "Socket closed?~%")
			     (close-connection ready))))))
	(loop for c in (cons server-socket (get-connections))
	      do (close-connection c))
	(stop-server)))))

(defun stop-server ()
  (setf *games* (make-hash-table :test 'equal))
  (setf *server-running* nil))

(defmethod handle-message (msg stream connection))

(defmethod handle-message ((msg g:game-action-message) stream connection)
  (declare (ignore stream connection))
  (let ((game (gethash (g:game-code msg) *games*)))
    (when game
      (push (g:game-action msg) (g:actions game)))))

(defmethod handle-message ((msg g:create-game-message) stream connection)
  (let ((game (make-instance 'online-game
			     :state (make-instance 'g:state
						   :left-player (g:player-name msg))
			     :connections (list connection)
			     :left-player-connection connection)))
    (setf (gethash (g:game-code game) *games*) game)
    (setf (gethash connection *connections*) game)
    (let ((response (make-instance 'g:game-created-message
				   :game-code (g:game-code game)
				   :player-paddle :left
				   :game-state (g:game-state game))))
      (cpk:encode response :stream stream))
    (force-output stream)))

(defmethod handle-message ((msg g:join-game-message) stream connection)
  (let ((game (gethash (g:game-code msg) *games*)))
    (if game
	(let ((left-player (g:state-left-player (g:game-state game)))
	      (right-player (g:state-right-player (g:game-state game))))
	  (if (or (not left-player) (not right-player))
	      (let ((paddle (cond (left-player :right) (right-player :left)))
		    (player (g:player-name msg)))
		(pushnew connection (online-game-connections game))
		(setf (gethash connection *connections*) game)

		(case paddle
		  (:left
		   (setf (g:state-left-player (g:game-state game)) player)
		   (setf (online-game-left-player-connection game) connection))
		  (:right
		   (setf (g:state-right-player (g:game-state game)) player)
		   (setf (online-game-right-player-connection game) connection)))

		(setf (g:game-state game)
		      (make-instance 'g:state
				     :left-player (g:state-left-player (g:game-state game))
				     :right-player (g:state-right-player (g:game-state game))))
		(g:random-launch-ball (g:game-state game))

		(let ((response (make-instance 'g:game-joined-message
					       :player-paddle paddle
					       :game-state (g:game-state game))))
		  (loop for con in (online-game-connections game)
			do (send-data-to-client response con))))
	      (send-message-to-client connection g:game-full)))
	(send-message-to-client connection g:game-does-not-exist))))

;; (start-server 54321)
;; (stop-server)

(defun main ()
  (format t "Starting server!~%")
  (unwind-protect (start-server 54321)
    (stop-server)))

